<?php

/**
 * Based on an example by Andac Aydin (aandac@gmx.de) from the PEL library (Copyright (C) 2007 Martin Geisler).
 * Copyright 2023 Sebastian Weiß, Jakob Berger
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program in the file COPYING; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301 USA
 */

/*
 * Low-level functions using PEL for reading and writing EXIF metadata to image files.
 * @since 1.1.0
 * @package geit
 */


use lsolesen\pel\PelEntryAscii;
use lsolesen\pel\PelEntryByte;
use lsolesen\pel\PelEntryRational;
use lsolesen\pel\PelExif;
use lsolesen\pel\PelIfd;
use lsolesen\pel\PelJpeg;
use lsolesen\pel\PelPng;
use lsolesen\pel\PelTag;
use lsolesen\pel\PelTiff;
use lsolesen\pel\PelWebP;

/**
 * Convert a decimal degree into degrees, minutes, and seconds.
 *
 * @param
 *            int the degree in the form 123.456. Must be in the interval
 *            [-180, 180].
 *
 * @return array a triple with the degrees, minutes, and seconds. Each
 *         value is an array itself, suitable for passing to a
 *         PelEntryRational. If the degree is outside the allowed interval,
 *         null is returned instead.
 */
function convertDecimalToDMS( $degree ) {
	if ( $degree > 180 || $degree < - 180 ) {
		return null;
	}

	$degree = abs( $degree ); // make sure number is positive
	// (no distinction here for N/S
	// or W/E).

	$seconds = $degree * 3600; // Total number of seconds.

	$degrees = floor( $degree ); // Number of whole degrees.
	$seconds -= $degrees * 3600; // Subtract the number of seconds
	// taken by the degrees.

	$minutes = floor( $seconds / 60 ); // Number of whole minutes.
	$seconds -= $minutes * 60; // Subtract the number of seconds
	// taken by the minutes.

	$seconds = round( $seconds * 100, 0 ); // Round seconds with a 1/100th
	// second precision.

	return [
		[
			$degrees,
			1,
		],
		[
			$minutes,
			1,
		],
		[
			$seconds,
			100,
		],
	];
}

function rational2float( array $rational ): float {
	if ( ( $count = count( $rational ) ) != 2 ) {
		throw new Exception( "Invalid rational! Wrong parameter count: $count! Expected: 2" );
	}

	return floatval( $rational[0] ) / $rational[1];
}


function get_image_location( $GPSLatitudeRef, $GPSLatitude, $GPSLongitudeRef, $GPSLongitude ) {
	$lat_degrees = count( $GPSLatitude ) > 0 ? rational2float( $GPSLatitude[0] ) : 0;
	$lat_minutes = count( $GPSLatitude ) > 1 ? rational2float( $GPSLatitude[1] ) : 0;
	$lat_seconds = count( $GPSLatitude ) > 2 ? rational2float( $GPSLatitude[2] ) : 0;

	$lon_degrees = count( $GPSLongitude ) > 0 ? rational2float( $GPSLongitude[0] ) : 0;
	$lon_minutes = count( $GPSLongitude ) > 1 ? rational2float( $GPSLongitude[1] ) : 0;
	$lon_seconds = count( $GPSLongitude ) > 2 ? rational2float( $GPSLongitude[2] ) : 0;

	$lat_direction = ( $GPSLatitudeRef == 'W' or $GPSLatitudeRef == 'S' ) ? - 1 : 1;
	$lon_direction = ( $GPSLongitudeRef == 'W' or $GPSLongitudeRef == 'S' ) ? - 1 : 1;

	$latitude  = $lat_direction * ( $lat_degrees + ( $lat_minutes / 60 ) + ( $lat_seconds / ( 60 * 60 ) ) );
	$longitude = $lon_direction * ( $lon_degrees + ( $lon_minutes / 60 ) + ( $lon_seconds / ( 60 * 60 ) ) );

	return array(
		'latitude'  => $latitude,
		'longitude' => $longitude,
	);

}


/**
 * Add GPS information to an image basic metadata.
 * Any old Exif data
 * is discarded.
 *
 * @param
 *            string the input filename.
 *
 * @param
 *            string the output filename. An updated copy of the input
 *            image is saved here.
 *
 * @param
 *            string image description.
 *
 * @param
 *            string user comment.
 *
 * @param
 *            string camera model.
 *
 * @param
 *            float longitude expressed as a fractional number of degrees,
 *            e.g. 12.345�. Negative values denotes degrees west of Greenwich.
 *
 * @param
 *            float latitude expressed as for longitude. Negative values
 *            denote degrees south of equator.
 *
 * @param
 *            float the altitude, negative values express an altitude
 *            below sea level.
 *
 * @param
 *            string the date and time.
 */


function validateCoordinates( $lat, $lon ) {
	$pattern = '/^-?\d{1,2}(?:\.\d+)?$/'; // Regular expression pattern for decimal coordinates

	if ( preg_match( $pattern, $lat ) && preg_match( $pattern, $lon ) ) {
		$latitude  = floatval( $lat );
		$longitude = floatval( $lon );

		if ( $latitude >= - 90 && $latitude <= 90 && $longitude >= - 180 && $longitude <= 180 ) {
			return true; // Valid coordinates
		}
	}

	return false; // Invalid coordinates
}

function addGpsInfo( $inputPath, $output, $longitude, $latitude, $author = '', $description = '', $check = false ) {
	$image = check_pel_type( $inputPath );
	if ( ! isset( $image ) or ! validateCoordinates( $latitude, $longitude ) ) {
		return false;
	}

	/*
	 * Create and add empty Exif data to the image (this throws away any
	 * old Exif data in the image).
	 */
	$image->clearExif();

	$exif = new PelExif();

	$gps_ifd = new PelIfd( PelIfd::GPS );

	/*
	 * Create and add TIFF data to the Exif data (Exif data is actually
	 * stored in a TIFF format).
	 */
	$tiff = new PelTiff();
	$exif->setTiff( $tiff );

	/*
	 * Create first Image File Directory and associate it with the TIFF
	 * data.
	 */
	$ifd0 = new PelIfd( PelIfd::IFD0 );
	$tiff->setIfd( $ifd0 );
	/*
	 * GPS data must be below the first IFD.
	 */
	$ifd0->addSubIfd( $gps_ifd );
	// moar stuff
	$inter_ifd = new PelIfd( PelIfd::INTEROPERABILITY );
	$ifd0->addSubIfd( $inter_ifd );

	// $ifd0->addEntry(new PelEntryAscii(PelTag::MODEL, $model));
	// $ifd0->addEntry(new PelEntryAscii(PelTag::DATE_TIME, $date_time));
	$ifd0->addEntry( new PelEntryAscii( PelTag::ARTIST, $author ) );
	$ifd0->addEntry( new PelEntryAscii( PelTag::IMAGE_DESCRIPTION, $description ) );

	/*
	 * The USER_COMMENT tag must be put in a Exif sub-IFD under the
	 * first IFD.
	 */
	// $exif_ifd = new PelIfd(PelIfd::EXIF);
	// $exif_ifd->addEntry(new PelEntryUserComment($comment));
	// $ifd0->addSubIfd($exif_ifd);
	$gps_ifd->addEntry( new PelEntryByte( PelTag::GPS_VERSION_ID, 2, 2, 0, 0 ) );

	/*
	 * Use the convertDecimalToDMS function to convert the latitude from
	 * something like 12.34� to 12� 20' 42"
	 */
	list( $hours, $minutes, $seconds ) = convertDecimalToDMS( $latitude );

	/* We interpret a negative latitude as being south. */
	$latitude_ref = ( $latitude < 0 ) ? 'S' : 'N';

	$gps_ifd->addEntry( new PelEntryAscii( PelTag::GPS_LATITUDE_REF, $latitude_ref ) );
	$gps_ifd->addEntry( new PelEntryRational( PelTag::GPS_LATITUDE, $hours, $minutes, $seconds ) );

	/* The longitude works like the latitude. */
	list( $hours, $minutes, $seconds ) = convertDecimalToDMS( $longitude );
	$longitude_ref = ( $longitude < 0 ) ? 'W' : 'E';

	$gps_ifd->addEntry( new PelEntryAscii( PelTag::GPS_LONGITUDE_REF, $longitude_ref ) );
	$gps_ifd->addEntry( new PelEntryRational( PelTag::GPS_LONGITUDE, $hours, $minutes, $seconds ) );
	$image->setExif( $exif );

	if ( $check == true ) {


		if ( file_put_contents( $output, $image->getBytes() ) == false ) {
			return false;
		} else {
			return true;
		}
	} else {
		return false;
	}
}

if ( ! function_exists( 'read_exif_data' ) ) {
	function read_exif_data( $file_path ) {
		$image = check_pel_type( $file_path );
		if ( ! isset( $image ) ) {
			return null;
		}
		$exif = $image->getExif();
		if ( ! isset( $exif ) ) {
			return null;
		}
		$ifd0    = $exif->getTiff()->getIfd();
		$tags    = [
			PelIfd::IFD0 => [
				'ARTIST'            => PelTag::ARTIST,
				'IMAGE_DESCRIPTION' => PelTag::IMAGE_DESCRIPTION,
			],
			PelIfd::GPS  => [
				'GPS_LATITUDE'      => PelTag::GPS_LATITUDE,
				'GPS_LATITUDE_REF'  => PelTag::GPS_LATITUDE_REF,
				'GPS_LONGITUDE'     => PelTag::GPS_LONGITUDE,
				'GPS_LONGITUDE_REF' => PelTag::GPS_LONGITUDE_REF,
			],
		];
		$results = [];
		foreach ( $tags as $ifdType => $fields ) {
			if ( $ifdType === PelIfd::IFD0 ) {
				$ifd = $ifd0;
			} else {
				$ifd = $ifd0->getSubIfd( $ifdType );
			}
			if ($ifd){
			foreach ( $fields as $fieldName => $fieldTag ) {
				if ( $ifd->getEntry( $fieldTag ) ) {
					$results[ $fieldName ] = isset( $ifd ) ? $ifd->getEntry( $fieldTag )->getValue() : null;
				} else {
					$results[ $fieldName ] = '';
				}
			}
		};
		}
		$gpsValues                         = [
			$results['GPS_LATITUDE'] ?? null,
			$results['GPS_LATITUDE_REF'] ?? null,
			$results['GPS_LONGITUDE'] ?? null,
			$results['GPS_LONGITUDE_REF'] ?? null,
		];
		$results['gps_coordinates_simple'] = null;
		if ( ! in_array( null, $gpsValues ) ) {
			[ $latitude, $latRef, $longitude, $longRef ] = $gpsValues;
			// $GPSLatitudeRef, $GPSLatitude, $GPSLongitudeRef, $GPSLongitude
			$results['gps_coordinates_simple'] = get_image_location( $latRef, $latitude, $longRef, $longitude );
		}

		return $results;
	}
}

function check_pel_type( $attachment ) {
	$extension = wp_check_filetype( $attachment )['ext'];

	if ( in_array( $extension, [ 'jpeg', 'jpg' ] ) ) {
		$image = new PelJpeg( $attachment );
	} elseif ( $extension == 'webp' ) {
		$image = new PelWebP( $attachment );
	} elseif ( $extension == 'png' ) {
		$image = new PelPng( $attachment );
	} else {
		throw new \Exception( "Unknown file extension encountered: $extension" );
	}

	return $image;
}